from dataclasses import dataclass
from typing import Any

from sampo.schemas.identifiable import Identifiable
from sampo.schemas.requirements import WorkerReq, EquipmentReq, MaterialReq, ConstructionObjectReq, ZoneReq
from sampo.schemas.resources import Material
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.utilities.serializers import custom_serializer


@dataclass
class WorkUnit(AutoJSONSerializable['WorkUnit'], Identifiable):
    """
    Class that describe vertex in graph (one work/task)
    """
    def __init__(self,
                 id: str,
                 model_name: dict[str, Any] | str,
                 worker_reqs: list[WorkerReq] = None,
                 equipment_reqs: list[EquipmentReq] = None,
                 material_reqs: list[MaterialReq] = None,
                 object_reqs: list[ConstructionObjectReq] = None,
                 zone_reqs: list[ZoneReq] = None,
                 description: str = '',
                 group: str = 'main project',
                 priority: int = 1,
                 is_service_unit: bool = False,
                 volume: float = 0,
                 display_name: str = "",
                 workground_size: int = 100):
        """
        :param model_name: dict with information that describes type of work for resource model.
                           In minimal it should contain 'granular_name' and 'measurement' entries.
                           `str` model_type is equal to {'granular_name': your_str_value, 'measurement': 'unit'}
        :param worker_reqs: list of required professions (i.e. workers)
        :param equipment_reqs: list of required equipment
        :param material_reqs: list of required materials (e.g. logs, stones, gravel etc.)
        :param object_reqs: list of required objects (e.g. electricity, pipelines, roads)
        :param zone_reqs: list of required zone statuses (e.g. opened/closed doors, attached equipment, etc.)
        :param description: the description. It is useful, for example, to show it on visualization
        :param group: union block of works
        :param is_service_unit: service units are additional vertexes
        :param volume: scope of work
        :param display_name: name of work
        """
        if isinstance(model_name, str):
            model_name = {'granular_name': model_name}
        if 'measurement' not in model_name:
            model_name['measurement'] = 'unit'

        self.model_name = model_name

        super(WorkUnit, self).__init__(id, 'dummy')

        if material_reqs is None:
            material_reqs = []
        if object_reqs is None:
            object_reqs = []
        if equipment_reqs is None:
            equipment_reqs = []
        if worker_reqs is None:
            worker_reqs = []
        if zone_reqs is None:
            zone_reqs = []
        self.worker_reqs = worker_reqs
        self.equipment_reqs = equipment_reqs
        self.object_reqs = object_reqs
        self.material_reqs = material_reqs
        self.zone_reqs = zone_reqs
        self.description = description
        self.group = group
        self.is_service_unit = is_service_unit
        self.volume = float(volume)
        self.display_name = display_name if display_name else model_name['granular_name']
        self.priority = priority

    def __del__(self):
        for attr in self.__dict__.values():
            del attr

    def need_materials(self) -> list[Material]:
        return [req.material() for req in self.material_reqs]

    @custom_serializer('worker_reqs')
    @custom_serializer('zone_reqs')
    @custom_serializer('material_reqs')
    def serialize_serializable_list(self, value) -> list:
        """
        Return serialized list of values.
        Values should be serializable.

        :param value: list of values
        :return: list of serialized values
        """
        return [t._serialize() for t in value]

    @classmethod
    @custom_serializer('material_reqs', deserializer=True)
    def material_reqs_deserializer(cls, value) -> list[MaterialReq]:
        """
        Get list of material requirements

        :param value: serialized list of material requirements
        :return: list of material requirements
        """
        return [MaterialReq._deserialize(wr) for wr in value]

    @classmethod
    @custom_serializer('worker_reqs', deserializer=True)
    def worker_reqs_deserializer(cls, value) -> list[WorkerReq]:
        """
        Get list of worker requirements

        :param value: serialized list of work requirements
        :return: list of worker requirements
        """
        return [WorkerReq._deserialize(wr) for wr in value]

    @classmethod
    @custom_serializer('zone_reqs', deserializer=True)
    def zone_reqs_deserializer(cls, value) -> list[ZoneReq]:
        """
        Get list of worker requirements

        :param value: serialized list of work requirements
        :return: list of worker requirements
        """
        return [ZoneReq._deserialize(wr) for wr in value]

    @classmethod
    @custom_serializer('material_reqs', deserializer=True)
    def material_reqs_deserializer(cls, value) -> list[MaterialReq]:
        """
        Get list of material requirements

        :param value: serialized list of material requirements
        :return: list of material requirements
        """
        return [MaterialReq._deserialize(wr) for wr in value]

    def __getstate__(self):
        # custom method to avoid calling __hash__() on GraphNode objects
        return self._serialize()

    def __setstate__(self, state):
        new_work_unit = self._deserialize(state)
        self.worker_reqs = new_work_unit.worker_reqs
        self.equipment_reqs = new_work_unit.equipment_reqs
        self.object_reqs = new_work_unit.object_reqs
        self.material_reqs = new_work_unit.material_reqs
        self.zone_reqs = new_work_unit.zone_reqs
        self.id = new_work_unit.id
        self.model_name = new_work_unit.model_name
        self.is_service_unit = new_work_unit.is_service_unit
        self.volume = new_work_unit.volume
        self.group = new_work_unit.group
        self.display_name = new_work_unit.display_name
        self.priority = new_work_unit.priority



from collections.abc import Iterable
from typing import Union, TypeVar, Callable

T = TypeVar('T')


def flatten(xs: Iterable[Union[list[T], T]]) -> Iterable[T]:
    """
    Returns a generator which should flatten any heterogeneous iterable

    :param xs:
    :return:
    """
    for x in xs:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x


K = TypeVar('K')
V = TypeVar('V')


def build_index(items: Iterable[T], key_getter: Callable[[T], K], value_getter: Callable[[T], V] = lambda x: x) \
        -> dict[K, V]:
    """
    :param items: an iterable to index
    :param key_getter: a function that should retrieve index key from item
    :param value_getter: a function that should retrieve index value from item
    :return: dictionary that represents built index given by `key_getter` function
    """
    return {key_getter(item): value_getter(item) for item in items}


def reverse_dictionary(dictionary: dict[K, V]) -> dict[V, K]:
    return {value: key for key, value in dictionary.items()}


def first(the_iterable, condition=lambda x: True):
    for i in the_iterable:
        if condition(i):
            return i



from typing import Callable

import numpy as np

from sampo.scheduler.utils import WorkerContractorPool
from sampo.schemas.contractor import Contractor
from sampo.schemas.exceptions import NoSufficientContractorError
from sampo.schemas.requirements import WorkerReq
from sampo.schemas.resources import Worker
from sampo.schemas.schedule_spec import WorkSpec
from sampo.schemas.time import Time


def get_worker_borders(agents: WorkerContractorPool, contractor: Contractor, work_reqs: list[WorkerReq]) \
        -> tuple[np.ndarray, np.ndarray, list[Worker]]:
    """
    Define for each job each type of workers the min and max possible number of workers.
    For max number of workers, max is defined as a minimum from max possible numbers
    at all and max possible for a current job.
    
    :param agents: from all projects
    :param contractor:
    :param work_reqs:
    :return:
    """
    n = len(work_reqs)
    min_worker_team = np.zeros(n, dtype=int)
    max_worker_team = np.zeros(n, dtype=int)
    workers: list[Worker] = []

    for i, req in enumerate(work_reqs):
        w = agents[req.kind].get(contractor.id, None)

        if w is None or w.count < req.min_count:
            # can't satisfy requirements, return empty to say it
            return [], [], []

        min_worker_team[i] = req.min_count
        max_worker_team[i] = min(req.max_count, w.count)
        workers.append(w)

    return min_worker_team, max_worker_team, workers


def run_contractor_search(contractors: list[Contractor],
                          work_spec: WorkSpec,
                          runner: Callable[[Contractor], tuple[Time, Time, list[Worker]]]) \
        -> tuple[Time, Time, Contractor, list[Worker]]:
    """
    Performs the best contractor search.
    
    :param contractors: contractors' list
    :param runner: a runner function, should be inner of the calling code.
        Calculates Tuple[start time, finish time, worker team] from given contractor object.
    :return: start time, finish time, the best contractor, worker team with the best contractor
    """
    # TODO Parallelize

    # optimization metric
    best_finish_time = Time.inf()
    best_contractor = None
    # heuristic: if contractors' finish times are equal, we prefer smaller one
    best_contractor_size = float('inf')

    contractors = work_spec.filter_contractors(contractors)

    for contractor in contractors:
        start_time, finish_time, worker_team = runner(contractor)
        contractor_size = sum(w.count for w in contractor.workers.values())

        if not finish_time.is_inf() and (finish_time < best_finish_time or
                                         (finish_time == best_finish_time and contractor_size < best_contractor_size)):
            best_finish_time = finish_time
            best_contractor = contractor
            best_contractor_size = contractor_size

    if best_contractor is None:
        raise NoSufficientContractorError(f'There is no contractor that can satisfy given search; contractors: '
                                          f'{contractors}')

    best_start_time, best_finish_time, best_worker_team = runner(best_contractor)

    return best_start_time, best_finish_time, best_contractor, best_worker_team



from functools import partial
from operator import itemgetter

from sampo.scheduler.utils import get_head_nodes_with_connections_mappings_nodes
from sampo.scheduler.utils.time_computaion import work_priority, calculate_working_time_cascade, \
    calculate_scheduled_time_cascade
from sampo.schemas import GraphNode, WorkTimeEstimator, ScheduledWork


def ford_bellman(nodes: list[GraphNode],
                 weights: dict[str, float],
                 node_id2parent_ids: dict[str, set[str]]) -> dict[str, float]:
    """
    Runs heuristic ford-bellman algorithm for given graph and weights.
    """
    path_weights: dict[str, float] = {node.id: 0 for node in nodes}
    # cache graph edges
    edges: list[tuple[str, str, float]] = sorted([(finish, start.id, weights[finish])
                                                  for start in nodes
                                                  for finish in node_id2parent_ids[start.id]
                                                  if finish in path_weights])

    if not edges:
        return path_weights

    # for changes heuristic
    changed = False
    # run standard ford-bellman on reversed edges
    # optimize dict access to finish weight
    for i in range(len(nodes)):
        cur_finish = edges[0][0]
        cur_finish_weight = path_weights[cur_finish]
        # relax on edges
        for finish, start, weight in edges:
            # we are running through the equality class by finish node
            # so if it changes renew the parameters of current equality class
            if cur_finish != finish:
                path_weights[cur_finish] = cur_finish_weight
                cur_finish = finish
                cur_finish_weight = path_weights[cur_finish]
            new_weight = path_weights[start] + weight
            if new_weight < cur_finish_weight:
                cur_finish_weight = new_weight
                changed = True
        # if we were done completely nothing with actual minimum weights, the algorithm ends
        if not changed:
            break
        # we go here if changed = True
        # so the last equality class weight can be changed, save it
        path_weights[cur_finish] = cur_finish_weight
        # next iteration should start without change info from previous
        changed = False

    return path_weights


def critical_path_graph(nodes: list[GraphNode], work_estimator: WorkTimeEstimator) -> list[GraphNode]:
    weights = {node.id: -work_priority(node, calculate_working_time_cascade, work_estimator) for node in nodes}
    head_nodes, node_id2parent_ids, node_id2child_ids = get_head_nodes_with_connections_mappings_nodes(nodes)
    return _critical_path(head_nodes, weights, node_id2parent_ids, node_id2child_ids)


def critical_path_schedule(nodes: list[GraphNode], scheduled_works: dict[str, ScheduledWork]) -> list[GraphNode]:
    weights = {node.id: -work_priority(node, partial(calculate_scheduled_time_cascade, id2swork=scheduled_works), None) for node in nodes}
    head_nodes, node_id2parent_ids, node_id2child_ids = get_head_nodes_with_connections_mappings_nodes(nodes)
    return _critical_path(head_nodes, weights, node_id2parent_ids, node_id2child_ids)


def critical_path_schedule_lag_optimized(nodes_optimized: list[GraphNode],
                                         scheduled_works_optimized: dict[str, ScheduledWork],
                                         nodes: list[GraphNode]) -> list[GraphNode]:
    node_dict = {node.id: node for node in nodes}

    cp_optimized = critical_path_schedule(nodes_optimized, scheduled_works_optimized)
    cp = []
    for node in cp_optimized:
        while node.is_inseparable_parent():
            node = node.inseparable_son
        if node_dict[node.id] not in cp:
            cp.append(node_dict[node.id])
    return cp


def _critical_path(nodes: list[GraphNode],
                   weights: dict[str, float],
                   node_id2parent_ids: dict[str, set[str]],
                   node_id2child_ids: dict[str, set[str]]) -> list[GraphNode]:
    cum_weights = ford_bellman(nodes, weights, node_id2parent_ids)
    node_dict = {node.id: node for node in nodes}

    critical_path = []

    start_node_id = min(cum_weights.items(), key=itemgetter(1))[0]
    node = node_dict[start_node_id]

    while node.children:
        critical_path.append(node)

        # sorted_nodes = sorted(node.children, key=lambda x: cum_weights[x.id])
        # [(node, cum_weights[node.id]) for node in critical_path]
        # node = sorted_nodes[0]
        node = node_dict[min(node_id2child_ids[node.id], key=lambda x: cum_weights[x])]

    return critical_path



from operator import itemgetter

from sampo.scheduler.utils.critical_path import ford_bellman
from sampo.utilities.priority import extract_priority_groups_from_nodes
from sampo.scheduler.utils.time_computaion import work_priority, calculate_working_time_cascade
from sampo.schemas.graph import GraphNode
from sampo.schemas.time_estimator import WorkTimeEstimator


def prioritization_nodes(nodes: list[GraphNode],
                         node_id2parent_ids: dict[str, set[str]],
                         work_estimator: WorkTimeEstimator) -> list[GraphNode]:
    """
    Return ordered critical nodes.
    Finish time is depended on these ordered nodes.
    """
    if len(nodes) == 1:
        return nodes

    # inverse weights
    weights = {node.id: -work_priority(node, calculate_working_time_cascade, work_estimator)
               for node in nodes}

    path_weights = ford_bellman(nodes, weights, node_id2parent_ids)

    ordered_nodes = sorted(nodes, key=lambda node: path_weights[node.id])

    return ordered_nodes


def prioritization(head_nodes: list[GraphNode],
                   node_id2parent_ids: dict[str, set[str]],
                   node_id2child_ids: dict[str, set[str]],
                   work_estimator: WorkTimeEstimator) -> list[GraphNode]:
    """
    Return ordered critical nodes.
    Finish time is depended on these ordered nodes.
    """

    # form groups
    groups = extract_priority_groups_from_nodes(head_nodes)

    result = []
    for _, group in sorted(groups.items(), key=itemgetter(0)):
        result.extend(prioritization_nodes(group, node_id2parent_ids, work_estimator))

    return result



from collections import defaultdict
from copy import copy
from dataclasses import dataclass, field

from sampo.schemas import Contractor
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time
from sampo.schemas.types import WorkerName
from sampo.schemas.works import WorkUnit


@dataclass
class WorkSpec:
    """
    Here are the container for externally given terms, that the resulting `ScheduledWork` should satisfy.
    Must be used in schedulers.
    :param contractors: list of contractors' ids that can be assigned to this work.
                        Should match the global contractors list
    :param assigned_workers: predefined worker team (scheduler should assign this worker team to this work)
    :param assigned_time: predefined work time (scheduler should schedule this work with this execution time)
    :param is_independent: should this work be resource-independent, e.g. executing with no parallel users of
    its types of resources
    """
    contractors: set[str] = field(default_factory=set)
    assigned_workers: dict[WorkerName, int] = field(default_factory=dict)
    assigned_time: Time | None = None
    is_independent: bool = False

    def is_contractor_enabled(self, contractor_id: str) -> bool:
        return len(self.contractors) == 0 or contractor_id in self.contractors

    def filter_contractors(self, contractors: list[Contractor]) -> list[Contractor]:
        return [contractor for contractor in contractors if self.is_contractor_enabled(contractor.id)]


@dataclass
class ScheduleSpec:
    """
    Here is the container for externally given terms, that Schedule should satisfy.
    Must be used in schedulers.
    """

    _work2spec: dict[str, WorkSpec] = field(default_factory=lambda: defaultdict(WorkSpec))

    def set_exec_time(self, work: str | WorkUnit, time: Time) -> 'ScheduleSpec':
        if isinstance(work, WorkUnit):
            work = work.id
        self._work2spec[work].assigned_time = time
        return self

    def assign_workers_dict(self, work: str | WorkUnit, workers: dict[WorkerName, int]) -> 'ScheduleSpec':
        if isinstance(work, WorkUnit):
            work = work.id
        self._work2spec[work].assigned_workers = copy(workers)
        return self

    def assign_workers(self, work: str | WorkUnit, workers: list[Worker]) -> 'ScheduleSpec':
        if isinstance(work, WorkUnit):
            work = work.id
        self._work2spec[work].assigned_workers = {worker.name: worker.count for worker in workers}
        return self

    def assign_contractors(self, work: str | WorkUnit, contractors: set[str]) -> 'ScheduleSpec':
        if isinstance(work, WorkUnit):
            work = work.id
        self._work2spec[work].contractors = contractors
        return self

    def get_work_spec(self, work_id: str) -> WorkSpec:
        return self._work2spec[work_id]

    def __getitem__(self, item):
        return self.get_work_spec(item)
